
#include "RobboLito.h"

void PawnHashReset ()
{
  memset (PawnHash, 0, PawnHashSize * sizeof (typePawnEval));
}

int InitPawnHash (int mb)
{
  if (mb > 1024)
    mb = 1024;
  PawnHashSize = ((1 << BSR (mb)) << 20) / sizeof (typePawnEval);
  mb = (PawnHashSize * sizeof (typePawnEval)) >> 20;
  if (PawnHash)
    ALIGNED_FREE (PawnHash);
  MEMALIGN (PawnHash, 64, PawnHashSize * sizeof (typePawnEval));
  PawnHashReset ();
  return (mb);
}

void ResetPositionalGain ()
{
  int p, m;
#ifdef MULTIPLE_POS_GAIN
  int cpu;
  for (cpu = 0; cpu < MAX_CPUS; cpu++)
#endif
    for (p = 0; p < 0x10; p++)
      for (m = 0; m < 010000; m++)
#ifdef MULTIPLE_POS_GAIN
	MAX_POSITIONAL_GAIN[cpu][p][m] = 0;
#else
        MAX_POSITIONAL_GAIN[p][m] = 0;
#endif
}

void ResetHistory ()
{
  int pi, sq;
#ifdef MULTIPLE_HISTORY
  int cpu;
  for (cpu = 0; cpu < MAX_CPUS; cpu++)
#endif
    for (pi = 0; pi < 0x10; pi++)
      for (sq = A1; sq <= H8; sq++)
#ifdef MULTIPLE_HISTORY
	HISTORY[cpu][pi][sq] = 0x800;
#else
	HISTORY[pi][sq] = 0x800;
#endif
}

static void Banner ()
{
  printf ("%s verzione %s\n", EPONYM, VERSION);
#ifdef HAS_PREFETCH
  printf ("compiled with PREFETCH\n");
#endif
#ifdef HAS_POPCNT
  printf ("compiled with POPCNT\n");
#endif
#ifdef YUSUF_MULTICORE
  printf ("Capitalist multicore mode compiled\n");
#endif
#ifdef UTILITIES
  printf ("Mode with UTILITIES compiled\n");
#endif
#ifdef MINIMAL
  printf ("MINIMAL Mode compiled\n");
#endif
#ifdef CON_ROBBO_BUILD
  printf ("RobboBases enabled by compile\n");
#endif
}

int main ()
{
  Banner ();
  PawnHash = NULL;
  PawnHashSize = (1 << 16);
  memset ( (void*) ROOT_POSIZIONE0, 0, sizeof (typePOS));
  ROOT_POSIZIONE0->DYN_ROOT =
    malloc (MAXIMUM_PLY * sizeof(typeDYNAMIC));
  ROOT_POSIZIONE0->DYN = ROOT_POSIZIONE0->DYN_ROOT; 
  ROOT_POSIZIONE0->stop = FALSE;
  CPUS_SIZE = 1;
#ifdef YUSUF_MULTICORE
  rp_init ();
#endif
  InitPawnHash (2);
  InitHash (32);
  ResetHistory ();
  InitCaptureValues ();
  InitArrays ();
  InitPawns ();
  UCI_BishopPairScale = 45;
  UCI_PawnScale = 100;
  UCI_KnightScale = 300;
  UCI_BishopScale = 310;
  UCI_RookScale = 500;
  UCI_QueenScale = 950;
  InitMaterialValue ();
  InitStatic ();
  NewGame (ROOT_POSIZIONE0, TRUE);

#ifdef USER_SPLIT
  CUT_SPLIT_DEPTH = 16;
  ALL_SPLIT_DEPTH = 14;
  PV_SPLIT_DEPTH = 14;
  SPLIT_AT_CUT = TRUE;
#endif
  SEND_CURR_MOVE = FALSE;
  USE_ALTERNATIVE_TIME = FALSE;
  BUFFER_TIME = 0;
  UCI_PONDER = FALSE;
  VERIFY_NULL = TRUE;
  ROBBO_LOAD = FALSE;
  SEARCH_ROBBO_BASES = TRUE;
  MULTI_PV = 1;
  SEND_HASH = FALSE;
  ALLOW_INSTANT_MOVE = TRUE;
  strcpy (ROBBO_TRIPLE_DIR, "");
  strcpy (ROBBO_TOTAL_DIR, "");

#ifndef MINIMAL
  MULTI_CENTI_PAWN_PV = 65535;
  RANDOM_COUNT = 0;
  RANDOM_BITS = 1;
  ALWAYS_ANALYZE = FALSE;
#endif

#ifdef CON_ROBBO_BUILD
  total_iniz ();
#endif
  while (1)
    Input (ROOT_POSIZIONE0);
  return 0;
}
